# Read in input files
## Main model selection file
setwd("/hpc/group/tmolab/lnc14/fecunda") # change to location where the directory has been saved on the remote server
load(file="fecunda-vrr-modelsel-output.Rdata")

### unlist relevant elements
coef<-modelsel$coefs
vcmat<-modelsel$vcmat
fixed<-modelsel$mainfx
ran<-modelsel$ranfx
yr.var<-modelsel$yr.var
yr.k<-modelsel$yr.k

## Residual variances for sampling variance in size
load(file="fecunda-vrr-residual-variances.Rdata")

## Weibull parameters for describing size distributions in 1997
load(file="fecunda-size-dists-97.Rdata")

## Empirical population sizes across all years
load(file="fecunda-empirical-N-90-97.Rdata")

## Raw demographic data
demog.all<-read.csv('/allpops_long_cleaned_dormancy.csv')

# Load source functions}
source("fecunda_ipm_source_functions.R")

# Load library for multivariate normal distributions
if(!require('MASS')) install.packages('MASS') 
library("MASS")

# Define IPM parameters
## Variable lists to loop over
poplist<-names(coef)
npop<-length(poplist)
vrrlist<-names(coef[[1]])
residlist<-names(resids.all[[1]])
nyrs<-length(c(90:97))
ntrans<-nyrs-1
yrs<-c(90:96)

## Model resolution
n<-150 # number of matrix cells in IPM

##  IPM range
min.size<-0.9*min(demog.all$size1, na.rm=T) # evaluate IPM starting here
max.size<-1.1*max(demog.all[demog.all$size2<100, 'size2'], na.rm=T) # and ending here (excludes outliers >100 mm diam)
b<-min.size+c(0:n)*(max.size-min.size)/n # boundary

## Mesh points and step size
m<-0.5*(b[(1:n)]+b[2:(n+1)])
d<-m[2]-m[1] 
nmp<-length(m)

# Calculate stable size distributions across 3 state spaces
## Initialize looping values
v<-1
p<-1

## Initialize output
n.stable<-matrix(NA, nrow=nmp*3, ncol=npop)
colnames(n.stable)<-poplist

stable.prop.dorm<-rep(NA, npop)
names(stable.prop.dorm)<-poplist
stable.prop.bolter<-stable.prop.dorm
stable.prop.normal<-stable.prop.dorm

## Monitor progress
t0<-Sys.time() # start timer
stop=FALSE

## Begin calculations

### loop over populations
for(p in 1:npop){
  # subset data to current pop (for parallel runs, this is determined from taskID)
  coef.pop<-coef[[p]]
  vcov.pop<-vcmat[[p]]
  fixed.pop<-fixed[[p]]
  ran.pop<-ran[[p]]
  yr.var.pop<-yr.var[[p]]
  yr.k.pop<-yr.k[[p]]
  resids.pop<-resids.all[[p]]

  # outside of bootstrapping loop, calculate mean matrix for stable size dist
  ## make dataframe for mean coefficients
  coef.mean<-data.frame(matrix(NA, nrow=length(vrrlist)*ntrans, ncol=7))
  colnames(coef.mean)<-c("vrr", "yr", '(Intercept)', 'size1', 'I(size1^2)', 'predorm', 'SD')
  coef.mean$vrr<-rep(vrrlist, each=ntrans)
  coef.mean$yr<-rep(c(90:96), length(vrrlist))

  ## assign mean VRR coefs from best-fit models
  v<-1
  for(v in 1:length(vrrlist)){
    coef.vrr<-coef.pop[[v]]
    fixed.vrr<-fixed.pop[[v]]
    ran.vrr<-ran.pop[[v]]
  
    # check dimensions of coefs; if a year is missing, take the average across years
    paste.coefs<-names(coef.vrr)
    y<-1
    for(y in 1:length(yrs)){
      if(rownames(coef.vrr)[y]%in%yrs){
        coef.mean[coef.mean$vrr==vrrlist[v]&coef.mean$yr==yrs[y], paste.coefs]<-coef.vrr[y,]
      }else{
        coef.mean[coef.mean$vrr==vrrlist[v]&coef.mean$yr==yrs[y], paste.coefs]<-colMeans(coef.vrr)
      }
    }
  }

  # add in SDs for size measurements
  r<-1
  for(r in 1:length(residlist)){
    mean.var<-resids.pop[[r]]$resid.var
    coef.mean[coef.mean$vrr==residlist[r], 'SD']<-sqrt(mean.var)
  }

  # replace NAs in mean coef dataframe with 0s
  coef.mean[is.na(coef.mean)]<-0

  # build annual matrices
  avg.kernels<-array(NA, c(nmp*3, nmp*3, ntrans))

  y<-1
  for(y in 1:length(yrs)){
    yr.now<-yrs[y]
  
    # define VRRs using mean coefficients from current year
    coef.use<-coef.mean[coef.mean$yr==yr.now,]
  
    # save this year's kernel (K; output of build_ipm function) to array of mean kernels
    avg.kernels[,,y]<-build_ipm()
  
  } # end year loop

  # calculate mean kernel across all years
  mean.K<-apply(avg.kernels, c(1,2), mean)

  # calculate stable size distribution for the mean kernel
  w.eigen<-Re(eigen(mean.K)$vectors[,1])
  n.stable[,p]<-(w.eigen/sum(w.eigen))

  # calculate the proportion of dormant individuals and bolters in the stable size distribution
  prop.dorm<-sum(n.stable[(nmp+1):(2*nmp), p])
  stable.prop.dorm[p]<-prop.dorm

  prop.bolter<-sum(n.stable[((2*nmp)+1):(3*nmp), p])
  stable.prop.bolter[p]<-prop.bolter

  prop.normal<-1-(prop.dorm+prop.bolter)
  stable.prop.normal[p]<-prop.normal

  ## initialize size distributions for use later
  ### define starting size distribution (from empirical 1997 dist)
  shape<-size.params.97[[p]][["shape"]]
  scale<-size.params.97[[p]][["scale"]]
  sizes<-dweibull(m, shape=shape, scale=scale)
  sizes<-sizes/sum(sizes)
  
  ### scale across state spaces (proportions in each state from stable size dist)
  n0<-matrix(NA,3*nmp,1)
  n0<-c((sizes*stable.prop.normal[[p]]), (sizes*stable.prop.dorm[[p]]), (sizes*stable.prop.bolter[[p]]))

  # Print progress report
  message(paste("Stable size distribution and mean kernels for population ", p, " computed. Elapsed time: ", round(difftime(Sys.time(),t0),4), " ", units(difftime(Sys.time(),t0)), sep=""))
  
  # Save population-specific outputs
  ## Array of annual kernels - mean coefs from each year
  save(avg.kernels, file=paste("fecunda-avg-Ks-peryear-pop", p, ".rda", sep=""))

  ## Mean kernel across all years
  write.csv(mean.K, file=paste("fecunda-mean-K-allyrs-pop", p, ".csv", sep=""), row.names=F)
  
  ## Stable size distribution 
  stable.dist.save<-data.frame(cbind(c(rep("normal", nmp), rep("dormant", nmp), rep("bolter", nmp)), c(m, m, m), n0))
  colnames(stable.dist.save)<-c("state", "size", "N")
  write.csv(stable.dist.save, file=paste("fecunda-stable-size-dist-pop", p, ".csv", sep=""), row.names=F)
  
  # Print progress report
  message(paste("Output saved to:", getwd(), sep=""))
  
} # end population loop
  