# Determine array task for cluster run (parallel across 5,000 bootstraps)
taskID<-as.integer(Sys.getenv('SLURM_ARRAY_TASK_ID')) # comment this out if running locally

# Set up environment

## Define working directory to navigate from, i.e., where the subdirectory is located on the remote (or local) server
setwd("/hpc/group/tmolab/lnc14/fecunda")

## Define SCRATCH directory to store temporary files (output will be one file per pop per bootstrap, which will then be combined post-hoc)
SCRATCH <- '/work/lnc14/fecunda'

# Read in input files
## Main model selection file
load(file="fecunda-vrr-modelsel-output.Rdata")

### unlist relevant elements
coef<-modelsel$coefs
vcmat<-modelsel$vcmat
fixed<-modelsel$mainfx
ran<-modelsel$ranfx
yr.var<-modelsel$yr.var
yr.k<-modelsel$yr.k

## Residual variances for sampling variance in size
load(file="fecunda-vrr-residual-variances.Rdata")

## Weibull parameters for describing size distributions in 1997
load(file="fecunda-size-dists-97.Rdata")

## Empirical population sizes across all years
load(file="fecunda-empirical-N-90-97.Rdata")

## Raw demographic data
demog.all<-read.csv('allpops_long_cleaned_dormancy.csv')

# Load source functions}
source("fecunda_ipm_source_functions.R")

# Load library for multivariate normal distributions
if(!require('MASS')) install.packages('MASS') 
library("MASS")

# Load dplyr for subsetting output
if(!require('dplyr')) install.packages('dplyr') 
library('dplyr')

# Define IPM parameters
## Variable lists to loop over
poplist<-names(coef)
npop<-length(poplist)
vrrlist<-names(coef[[1]])
residlist<-names(resids.all[[1]])
nyrs<-length(c(90:97))
ntrans<-nyrs-1
yrs<-c(90:96)

## Model resolution
iters<-50000 # iterations to calculate lamstoc within a bootstrap run
burnin<-0
nboot<-1 # number bootstrap runs PER JOB (i.e. number of different samples from VRR coef distributions... can do more jobs with fewer boots, or fewer jobs with more boots, depending on computing time)
n<-150 # number of matrix cells in IPM
nproj<-1000 # number of near-term population projections per bootstrap

##  IPM range
min.size<-0.9*min(demog.all$size1, na.rm=T) # evaluate IPM starting here
max.size<-1.1*max(demog.all[demog.all$size2<100, 'size2'], na.rm=T) # and ending here (excludes outliers >100 mm diam)
b<-min.size+c(0:n)*(max.size-min.size)/n # boundary

## Mesh points and step size
m<-0.5*(b[(1:n)]+b[2:(n+1)])
d<-m[2]-m[1] 
nmp<-length(m)

# Run Simulations
## Initialize looping values
boot<-1
iter<-1
v<-1
p<-1 

## Initialize output
### asymptotic dynamics
lam.boot<-matrix(NA, nrow=npop, ncol=nboot)
row.names(lam.boot)<-poplist

### stable size distributions across 3 state spaces
n.stable<-matrix(NA, nrow=nmp*3, ncol=npop)
colnames(n.stable)<-poplist

stable.prop.dorm<-rep(NA, npop)
names(stable.prop.dorm)<-poplist
stable.prop.bolter<-stable.prop.dorm
stable.prop.normal<-stable.prop.dorm

### population projections
#### make main dataframe for tmp output
proj<-data.frame(matrix(NA, nrow=npop*nboot*nproj, ncol=33))
colnames(proj)<-c("Pop", "Boot", "Projection", c(1990:2019)-1900)
proj$Pop<-rep(poplist, each=nboot*nproj)
proj$Boot<-rep(1:nboot, each=nproj)
proj$Projection<-rep(c(1:nproj))
fyrs<-c(1998:2019)
nfuture<-length(fyrs)

## Monitor progress
t0<-Sys.time() # start timer
stop=FALSE

## Begin

### loop over populations
for(p in 1:npop){
  
  #### populate empirical data for 1990-1997
  N.pop<-pop.sizes[[p]]
  proj[proj$Pop==poplist[p], names(N.pop)]<-rep(N.pop, each=nproj*nboot)
  
  #### make small dataframes for saved output (just summary statistics)
  proj.mean.save<-proj[1,]
  proj.mean.save[,]<-NA
  
  proj.sd.save<-proj.mean.save
  
  proj.median.save<-proj.mean.save
  
  proj.ext.save<-c(NA)
  
  # subset data to current pop (for parallel runs, this is determined from taskID)
  coef.pop<-coef[[p]]
  vcov.pop<-vcmat[[p]]
  fixed.pop<-fixed[[p]]
  ran.pop<-ran[[p]]
  yr.var.pop<-yr.var[[p]]
  yr.k.pop<-yr.k[[p]]
  resids.pop<-resids.all[[p]]
  
  # read in stable size vector for this pop
  n0<-read.csv(paste('./fecunda-stable-size-dist-pop', p, '.csv', sep=""))
  
  ### scale to empirical population size in 1997 for projections
  np0<-n0$N*N.pop[["97"]]

  # Simulate projections for population viability and lamstoc analyses 

  # loop over bootstrap runs
  for(boot in 1:nboot){
    if(stop==TRUE){break}
  
    # Within this bootstrap run, sample VRR coefs from distributions
    ## make dataframe to store sampled coefficients
    coef.sample<-data.frame(matrix(NA, nrow=length(vrrlist)*ntrans, ncol=7))
    colnames(coef.sample)<-c("vrr", "yr", '(Intercept)', 'size1', 'I(size1^2)', 'predorm', 'SD')
    coef.sample$vrr<-rep(vrrlist, each=ntrans)
    coef.sample$yr<-rep(c(90:96), length(vrrlist))
  
    ## sample coefficients for this bootstrap run
    for(v in 1:length(vrrlist)){
    
      # subset to current vital rate
      coef.vrr<-coef.pop[[v]]
      vc.vrr<-vcov.pop[[v]]
      fixed.vrr<-fixed.pop[[v]]
      ran.vrr<-ran.pop[[v]]
      yr.var.vrr<-yr.var.pop[[v]]
      k<-yr.k.pop[[v]]
    
      # sample main effects coefficients
      main.sample<-mvrnorm(n=1, fixed.vrr, vc.vrr)
    
      # add sampled coefs into dataframe
      s<-1
      for(s in 1:length(main.sample)){
        paste.sample<-names(main.sample)[s]
        coef.sample[coef.sample$vrr==vrrlist[v], paste.sample]<-main.sample[s]
      }
    
      # sample random deviations for each year
      ## draw variance from the chisq distribution of year-to-year variance estimate given df of years
      var.boot<-yr.var.vrr*(rchisq(1,(k-1))/(k-1))
    
      ## retain rank order effects of specific years by obtaining their position in the CDF of the distribution of year effects
      yr.ranks<-pnorm(ran.vrr$`(Intercept)`, mean=0, sd=sqrt(yr.var.vrr))
    
      ## sample yearly deviations from normal distribution of year-to-year variance, retaining rank order defined above
      if(sum(yr.ranks)>1){
        yr.ranks<-yr.ranks/sum(yr.ranks)} # if years are tied, assign each equal probability
      yr.boot<-qnorm(yr.ranks, mean=0, sd=sqrt(var.boot))
    
      ### if there are >7 obs (i.e., for repro, obs. in each year), just use the first 7
      ifelse(length(yr.boot)>7, yr.boot.use<-yr.boot[1:7], yr.boot.use<-yr.boot)
    
      # add sampled random deviations in each year onto the sampled main intercept
      ## determine which  years are directly observed
      paste.years<-rownames(ran.vrr)
    
      ## for each year with direct observations, add each sampled ranef deviation onto the main intercept sample
      coef.sample[coef.sample$vrr==vrrlist[v]&coef.sample$yr%in%paste.years, '(Intercept)']<-yr.boot.use+main.sample[1]
    
      ## otherwise, add mean of sampled ranef deviations to sampled intercept in un-observed years
      coef.sample[coef.sample$vrr==vrrlist[v]&!(coef.sample$yr%in%paste.years), '(Intercept)']<-mean(yr.boot)+main.sample[1]
    }
  
    ## add in sampled SDs for size measurements
    r<-1
    for(r in 1:length(residlist)){
      if(stop==TRUE){break}
      sampled.var<-resids.pop[[r]]$resid.var*(rchisq(1, (resids.pop[[r]]$k-1))/(resids.pop[[r]]$k-1))
      coef.sample[coef.sample$vrr==residlist[r], 'SD']<-sqrt(sampled.var)}
  
    ## replace NAs in coef dataframe with 0s
    coef.sample[is.na(coef.sample)]<-0
  
    # build annual matrices
    ## make array to store output
    sampled.kernels<-array(NA, c(nmp*3, nmp*3, ntrans))
  
    y<-1
    for(y in 1:length(yrs)){
      if(stop==TRUE){break}
      yr.now<-yrs[y]
    
      # define VRRs using sampled coefficients from current environment
      coef.use<-coef.sample[coef.sample$yr==yr.now,]
    
      # save this annual kernel (K; output of build_ipm) to array of sampled kernels
      sampled.kernels[,,y]<-build_ipm()
    
    } # end year loop
  
    # estimate stochastic lambda from sequence of random annual matrices
    ## define sequence of sampled environments (different in every bootstrap run)
    env<-ceiling(ntrans*runif(iters))
  
    ## construct list to temporarily hold output of iterations
    lam.seq<-matrix(NA, 1, iters)
  
    ## start bootstrap runs using stable size distribution (scaled to sum of 1)
    n1<-n0$N
  
    for(iter in 1:iters){
      # iterate lambda using random draws of sampled annual matrices
      ## project forward using current annual matrix
      n2<-sampled.kernels[,,env[iter]]%*%n1  
    
      ## sum population size at forward time step
      N<-sum(n2)
    
      ## save lambda for this iteration (i.e. pop size, since sum(n1)==1)
      lam.seq[iter]<-N
    
      ## renormalize size distribution vector given past projection
      n1<-n2/N
      
    } # end iterations
    
    # calculate stochastic lambda from list of lambdas
    lam.use<-c(lam.seq[(burnin+1):length(lam.seq)])
    lamstoc<-exp(mean(log(lam.use)))
    
    if(is.na(lamstoc)){
      message("NA value for lamstoc; bootstraps ceased.")
      message(paste("Pop: ", poplist[p], "; bootstrap run: ", boot, sep=""))
      stop=TRUE
      break
    }
    
    # save stochastic lambda in dataframe of bootstrap trials
    lam.boot[p,boot]<-lamstoc
    
    # project forward 22 years 1000 times with each set of sampled kernels
    pr<-1
    for(pr in 1:nproj){
      env.proj<-ceiling(ntrans*runif(nfuture))
      
      ## start population projections using empirical population size
      np1<-np0
      
      fyr<-1
      for(fyr in 1:nfuture){
        
        # project forward population size and distribution from the current sampled kernel
        np2<-sampled.kernels[,,env.proj[fyr]]%*%np1
        
        # sum projected population size at the next time step
        N.now<-round(sum(np2))
        
        # save pop size at this step of this projection
        proj[proj$Pop==poplist[p] & proj$Boot==boot & proj$Projection==pr, 11+fyr]<-N.now
        
        # save current pop size as starting point for next time step
        np1<-np2
      
      } # end future year sequence
      
    } # end replicate projections within each bootstrap run
    
    #### save "raw" trajectories for 1% of bootstrap runs (1 out of every 100 bootstraps)
    if(taskID==1|taskID%%100==0){
      write.csv(proj[proj$Pop==poplist[p],], paste(SCRATCH, "/projections/pop", p, "/raw/fecunda-N-raw-run", taskID, ".csv", sep=""), row.names=F)
    }
    
    #### save current VRR parameters to look at distributions of sampled values
    coef.write<-cbind(rep(NA, nrow(coef.sample)), coef.sample)
    colnames(coef.write)[1]<-"pop"
    coef.write$pop<-poplist[p]
    write.csv(coef.write, paste(SCRATCH, "/sampled_params/pop", p, "/fecunda-bootstrapped-params-run", taskID, ".csv", sep=""), row.names=F)
      
    #### make small dataframes for saved derived output
    proj.p<-proj[proj$Pop==poplist[p],]
    proj.N.mean<-proj.p[1,]
    proj.N.mean[,12:ncol(proj.N.mean)]<-NA
    
    proj.N.median<-proj.N.mean
      
    proj.N.sd<-proj.N.mean
      
    proj.ext<-proj.N.mean[,1:11]
      
    #### calculate mean, median, and SD of population size at each future time step
    fyr<-1
    for(fyr in 1:nfuture){
      proj.N.mean[1,11+fyr]<-mean(proj.p[, 11+fyr])
      proj.N.median[1,11+fyr]<-median(proj.p[,11+fyr])
      proj.N.sd[1,11+fyr]<-sd(proj.p[, 11+fyr])
    }
      
    #### calculate the proportion of projections that result in extinction
    proj.p$ext<-rep(NA, nrow(proj.p))
    for(row in 1:nrow(proj.p)){
      if(proj.p[row, ncol(proj)-1]>0){
        proj.p[row, 'ext']<-0}
      else(proj.p[row, 'ext']<-1)
    }
    
    proj.ext<-sum(proj.p$ext)/nrow(proj.p)
    
    # save just the summary statistics across the 1000 projections
    proj.mean.save<-rbind(proj.mean.save, proj.N.mean)
    
    proj.sd.save<-rbind(proj.sd.save, proj.N.sd)
    
    proj.median.save<-rbind(proj.median.save, proj.N.median)
    
    proj.ext.save<-rbind(proj.ext.save, proj.ext)
      
  } # end bootstrap loop
  
  # Save population-specific outputs
  
  ## Bootstrapped lambda values
  write.csv(lam.boot[poplist[p],], paste(SCRATCH, "/lamstoc/pop", p, "/fecunda-bootstrapped-lamstoc-run", taskID, '.csv', sep=""), row.names=T)
  
  ## Projected population sizes - mean
  proj.mean.save<-proj.mean.save[2:nrow(proj.mean.save),]
  write.csv(proj.mean.save, paste(SCRATCH, "/projections/pop", p, "/fecunda-N-means-run", taskID, ".csv", sep=""), row.names=F)
  
  ## Projected population sizes - sds
  proj.sd.save<-proj.sd.save[2:nrow(proj.sd.save),]
  write.csv(proj.sd.save, paste(SCRATCH, "/projections/pop", p, "/fecunda-N-sd-run", taskID, ".csv", sep=""), row.names=F)
  
  ## Projected population sizes - median
  proj.median.save<-proj.median.save[2:nrow(proj.median.save),]
  write.csv(proj.median.save, paste(SCRATCH, "/projections/pop", p, "/fecunda-N-medians-run", taskID, ".csv", sep=""), row.names=F)
  
  ## Extinction risk
  proj.ext.save<-proj.ext.save[2:nrow(proj.ext.save),]
  write.csv(proj.ext.save, paste(SCRATCH, "/extinction/pop", p, "/fecunda-pr-ext-run", taskID, ".csv", sep=""), row.names=F)
  
  # Print progress report
  message(paste('ALL SIMULATIONS FOR POPULATION', p, 'COMPLETE. Elapsed time:', round(difftime(Sys.time(), t0), 4), units(difftime(Sys.time(),t0))), sep="")
  
} # end population loop
