# This code defines the vital rate functions for the IPM
# It works by pulling the empirical or sampled coefficients for the particular vital rate, pop, and year
# from a dataset of all empirical or sampled coefficients.
# Because coefficients not retained after model selection are retained as 0s in the dataset, 
# they drop out of vital rate functions in which they are irrelevant.

### entering dormancy - bolters
dormB<-function(x, coef.use){
  linear<-c(coef.use[coef.use$vrr=="dormB", '(Intercept)']+
              coef.use[coef.use$vrr=="dormB", "size1"]*x+
              coef.use[coef.use$vrr=="dormB", "I(size1^2)"]*x^2)
  prob<-exp(linear)/(1+exp(linear))
  #prob[!is.finite(prob)]<-1
  return(prob)}

### entering dormancy - non-bolters
dormNB<-function(x, coef.use){
  linear<-c(coef.use[coef.use$vrr=="dormNB", '(Intercept)']+
              coef.use[coef.use$vrr=="dormNB", "size1"]*x+
              coef.use[coef.use$vrr=="dormNB", "I(size1^2)"]*x^2)
  prob<-exp(linear)/(1+exp(linear))
  #prob[!is.finite(prob)]<-1
  return(prob)}

### exiting dormancy
exit<-function(x, coef.use){
  linear<-c(coef.use[coef.use$vrr=="exit.dorm", '(Intercept)']+
              coef.use[coef.use$vrr=="exit.dorm", "predorm"]*x)
  prob<-exp(linear)/(1+exp(linear))
  #prob[!is.finite(prob)]<-1
  return(prob)}

### size upon exiting dormancy
exit.size<-function(y, x, coef.use){
  mean<-(coef.use[coef.use$vrr=="exit.dorm.size", '(Intercept)']+
           coef.use[coef.use$vrr=="exit.dorm.size", 'predorm']*x)
  sd<-coef.use[coef.use$vrr=="exit.dorm.size", 'SD']
  sizes<-dnorm(y, mean=mean, sd=sd)
  return(sizes)}

### bolter-ing
bolter<-function(x, coef.use){
  linear<-c(coef.use[coef.use$vrr=="bolter", '(Intercept)']+
              coef.use[coef.use$vrr=="bolter", 'size1']*x+
              coef.use[coef.use$vrr=="bolter", 'I(size1^2)']*x^2)
  prob<-exp(linear)/(1+exp(linear))
  #prob[!is.finite(prob)]<-1
  return(prob)}

### remaining a bolter
rebolter<-function(x, coef.use){
  linear<-(coef.use[coef.use$vrr=="rebolter", '(Intercept)']+
             coef.use[coef.use$vrr=="rebolter", 'prebolter']*x)
  prob<-exp(linear)/(1+exp(linear))
  #prob[!is.finite(prob)]<-1
  return(prob)}

### growth - bolters
grB<-function(y, x, coef.use){
  #### mean growth
  mean<-(coef.use[coef.use$vrr=="growthB", '(Intercept)']+
           coef.use[coef.use$vrr=="growthB", 'size1']*x+
           coef.use[coef.use$vrr=="growthB", 'I(size1^2)']*x^2)
  # mean[mean>max.size]<-max.size
  #### size-dependent variance in growth
  sd<-coef.use[coef.use$vrr=="growthB", "SD"]
  sd[sd<0]<-0 # change negative variance to 0 variance (alternative would be abs)
  sizes<-dnorm(y, mean=mean, sd=sd)
  return(sizes)}

### growth - non-bolters
grNB<-function(y, x, coef.use){
  #### mean growth
  mean<-(coef.use[coef.use$vrr=="growthNB", '(Intercept)']+
           coef.use[coef.use$vrr=="growthNB", 'size1']*x+
           coef.use[coef.use$vrr=="growthNB", 'I(size1^2)']*x^2)
  # mean[mean>max.size]<-max.size
  #### size-dependent variance in growth
  sd<-(coef.use[coef.use$vrr=="gr.residNB", '(Intercept)']+
         coef.use[coef.use$vrr=="gr.residNB", 'size1']*x+
         coef.use[coef.use$vrr=="gr.residNB", 'I(size1^2)']*x^2)
  sd[sd<0]<-0 # change negative variance to 0 variance (alternative would be abs)
  sizes<-dnorm(y, mean=mean, sd=sd)
  return(sizes)}

### pr(rep) - bolters
pr.repB<-function(x, coef.use){
  linear<-(coef.use[coef.use$vrr=="didrepB", '(Intercept)']+
             coef.use[coef.use$vrr=="didrepB", "prebolter"]*x)
  prob<-exp(linear)/(1+exp(linear))
  #prob[!is.finite(prob)]<-1
  return(prob)}

### pr(rep) - non-bolters
pr.repNB<-function(x, coef.use){
  linear<-(coef.use[coef.use$vrr=="didrepNB", '(Intercept)']+
             coef.use[coef.use$vrr=="didrepNB", "size1"]*x+
             coef.use[coef.use$vrr=="didrepNB", "I(size1^2)"]*x^2)
  prob<-exp(linear)/(1+exp(linear))
  #prob[!is.finite(prob)]<-1
  return(prob)}

### fruits - bolters
reproB<-function(x, coef.use){
  amt<-(coef.use[coef.use$vrr=="frtsB", '(Intercept)']+
          coef.use[coef.use$vrr=="frtsB", 'prebolter']*x)
  #amt[amt<0]<-0 # change negative estimates to 0
  #amt[amt>max.frts]<-max.frts # limit prediction of extremely high values
  return(amt)}

### fruits - non-bolters
reproNB<-function(x, coef.use){
  amt<-(coef.use[coef.use$vrr=="frtsNB", '(Intercept)']+
          coef.use[coef.use$vrr=="frtsNB", 'size1']*x+
          coef.use[coef.use$vrr=="frtsNB", 'I(size1^2)']*x^2)
  #amt[amt<0]<-0 # change negative estimates to 0
  #amt[amt>max.frts]<-max.frts # limit prediction of extremely high values
  return(amt)}

### survival - bolters
survB<-function(x, coef.use){
  linear<-(coef.use[coef.use$vrr=="survB", '(Intercept)']+
             coef.use[coef.use$vrr=="survB", 'prebolter']*x)
  prob<-exp(linear)/(1+exp(linear))
  return(prob)}

### survival - non-bolters
survNB<-function(x, coef.use){
  linear<-(coef.use[coef.use$vrr=="survNB", '(Intercept)']+
             coef.use[coef.use$vrr=="survNB", 'size1']*x+
             coef.use[coef.use$vrr=="survNB", 'I(size1^2)']*x^2)
  prob<-exp(linear)/(1+exp(linear))
  return(prob)}   

### recruitment
rec<-function(coef.use){
  linear<-(coef.use[coef.use$vrr=="rec.rate", '(Intercept)'])
  prob<-exp(linear)/(1+exp(linear))
  return(prob)}

### size distribution of new recruits
newsize<-function(y, coef.use){
  mean<-coef.use[coef.use$vrr=="rec.size", '(Intercept)']
  sd<-coef.use[coef.use$vrr=="rec.size", 'SD']
  sizes<-dlnorm(y, meanlog=mean, sdlog=sd)
  return(sizes)}

## tie together all repro VRRs - bolters
fec.fullB<-function(x, coef.use){
  pr.repB(x, coef.use)*reproB(x, coef.use)*rec(coef.use)}

## tie together all repro VRRs - non-bolters
fec.fullNB<-function(x, coef.use){
  pr.repNB(x, coef.use)*reproNB(x, coef.use)*rec(coef.use)}

build_ipm<-function(){
  
  ## construct IPM ########################################
  ### basic survival matrix
  S<-matrix(0, nrow=nmp, ncol=nmp)
  
  ### make bolter vs non bolter survival submatrices
  Sb=Snb=S
  
  #### apply survival probabilities across matrix
  Sb[,]<-rep(survB(x=m, coef.use=coef.use), each=nmp)  
  # OR:   Sb<-matrix(1,nmp,1) %*% matrix(survB(m,coef.use),1,nmp)
  
  Snb[,]<-rep(survNB(x=m, coef.use = coef.use), each=nmp)
  
  #### make survival megamatrix
  megaS<-cbind(rbind(Snb, Snb, Snb), rbind(Snb, Snb, Snb), rbind(Sb, Sb, Sb))
  
  ### state space transition matrix
  T11=T12=T13=T21=T22=T23=T31=T32=T33=matrix(NA, nmp, nmp)
  
  #### top left, normal to normal
  T11[,]<-rep(c(c(1-dormNB(x=m, coef.use=coef.use))*c(1-bolter(x=m, coef.use=coef.use))), each=nmp)
  
  #### top middle, dormant to normal
  T12[,]<-rep(c(c(exit(x=m, coef.use=coef.use))*c(1-bolter(x=m, coef.use=coef.use))), each=nmp)
  
  #### top right, bolter to normal
  T13[,]<-rep(c(c(1-rebolter(x=m, coef.use=coef.use))*c(1-dormB(x=m, coef.use=coef.use))), each=nmp)
  
  #### middle left, normal to dormant
  T21[,]<-rep(c(dormNB(x=m, coef.use=coef.use)), each=nmp)
  
  #### middle middle, dormant to dormant
  T22[,]<-rep(c(1-exit(x=m, coef.use=coef.use)), each=nmp)
  
  #### middle right, bolter to dormant
  T23[,]<-rep(c(c(1-rebolter(x=m, coef.use=coef.use))*c(dormB(x=m, coef.use=coef.use))), each=nmp)
  
  #### bottom left, normal to bolter
  T31[,]<-rep(c(c(1-dormNB(x=m, coef.use=coef.use))*c(bolter(x=m, coef.use=coef.use))), each=nmp)
  
  #### bottom middle, dormant to bolter
  T32[,]<-rep(c(c(exit(x=m, coef.use=coef.use))*c(bolter(x=m, coef.use=coef.use))), each=nmp)
  
  #### bottom right, bolter to bolter
  T33[,]<-rep(c(rebolter(x=m, coef.use=coef.use)), each=nmp)
  
  ### combine submatrices for state transition megamatrix
  megaT<-rbind(cbind(T11, T12, T13), 
               cbind(T21, T22, T22),
               cbind(T31, T32, T33))
  
  ### size/growth
  G11=G12=G13=Gx=diag(nmp)
  
  for(i in 1:nmp){
    for(j in 1:nmp){
      G11[i,j]<-grNB(y=m[i], x=m[j], coef.use=coef.use)*d # non dorm growth
      G12[i,j]<-exit.size(y=m[i], x=m[j], coef.use=coef.use)*d # exit dorm size
      G13[i,j]<-grB(y=m[i], x=m[j], coef.use=coef.use)*d}} # exit bolter size
  
  # renormalize to avoid eviction
  for(j in 1:nmp){
    sumG11<-sum(G11[,j])
    G11[,j]<-G11[,j]/sumG11
    
    sumG12<-sum(G12[,j])
    G12[,j]<-G12[,j]/sumG12
    
    sumG13<-sum(G13[,j])
    G13[,j]<-G13[,j]/sumG13
    
    # the 5 lines below can be a filter for a maximum tolerable amount of eviction. I have commented it out for now just to get the rest of the code running.
    # if(sumG11<.75 | sumG12<.75 | sumG13<.75){
    #   print("Growth function creating eviction of >25%. Iterations ceased.")
    #   print(paste("Pop: ", poplist[p], "; bootstrap run: ", boot, "; iteration: ", iter, "; i=", i, "; j=", j, sep=""))
    #   stop=TRUE
    #   break}
  }
  
  #### bind sub-matrices together
  megaG<-rbind(cbind(G11, G12, G13), 
               cbind(Gx, Gx, Gx), 
               cbind(Gx, Gx, Gx)) #Gx are identity matrices 
  
  ### check for negative growth values
  if(min(megaG)<0 | sum(is.na(megaG)>0)){
    print("Negative growth probability or NA in megaG; iterations ceased.")
    print(paste("Pop: ", poplist[p], "; bootstrap run: ", boot, "; iteration: ", iter, sep=""))
    stop=TRUE
    break
  }
  
  ## multiply together Surv, Transition, and Growth to make P matrix
  megaP<-megaS*megaT*megaG
  
  ## F matrices
  Fb = Fnb = Fx = matrix(0,nmp,nmp)
  
  ### make vectors of recruit numbers per size class
  recnum.b<-matrix(fec.fullB(x=m, coef.use=coef.use), 1, nmp)
  recnum.nb<-matrix(fec.fullNB(x=m, coef.use=coef.use), 1, nmp)
  
  ### make vector of recruit sizes
  recsize<-newsize(y=m, coef.use=coef.use)*d
  
  ### check for eviction in sizes
  if(sum(recsize)<.8){   
    print("Recruit size function creating eviction of >20%. Iterations ceased.")
    print(paste("Pop: ", poplist[p], "; bootstrap run: ", boot, "; iteration: ", iter, sep=""))
    stop=TRUE
    break}
  
  recsize<-matrix(recsize/sum(recsize), nmp, 1)
  
  ### multiply rows and columns to populate F kernels
  Fb<-recsize%*%recnum.b
  Fnb<-recsize%*%recnum.nb
  
  ### combine into megaF 
  megaF<-rbind(cbind(Fnb, Fnb, Fnb),cbind(Fx, Fx, Fx), cbind(Fb, Fb, Fb)) 
  megaF<-megaF*megaT # scale across state space by transition matrix
  
  ## sum full kernel
  return(megaP+megaF) # == K
  
  # save this annual kernel to array of sampled kernels
  # sampled.kernels[,,y]<-K
}
